/*
 * Copyright (C) 2009 by Matthias Ringwald
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MATTHIAS RINGWALD AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MATTHIAS
 * RINGWALD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#import <BTstack/BTstackManager.h>

#import <btstack/btstack.h>
#import <BTDevice.h>

// Control field values      bit no.       1 2 3 4 5   6 7 8
#define BT_RFCOMM_SABM       0x3F       // 1 1 1 1 P/F 1 0 0
#define BT_RFCOMM_UA         0x73       // 1 1 0 0 P/F 1 1 0
#define BT_RFCOMM_DM         0x0F       // 1 1 1 1 P/F 0 0 0
#define BT_RFCOMM_DM_PF      0x1F
#define BT_RFCOMM_DISC       0x53       // 1 1 0 0 P/F 0 1 1
#define BT_RFCOMM_UIH        0xEF       // 1 1 1 1 P/F 1 1 1
#define BT_RFCOMM_UIH_PF     0xFF

// Multiplexer message types 
#define BT_RFCOMM_PN_CMD     0x83
#define BT_RFCOMM_PN_RSP     0x81
#define BT_RFCOMM_TEST_CMD   0x23
#define BT_RFCOMM_TEST_RSP   0x21
#define BT_RFCOMM_FCON_CMD   0xA3
#define BT_RFCOMM_FCON_RSP   0xA1
#define BT_RFCOMM_FCOFF_CMD  0x63
#define BT_RFCOMM_FCOFF_RSP  0x61
#define BT_RFCOMM_MSC_CMD    0xE3
#define BT_RFCOMM_MSC_RSP    0xE1
#define BT_RFCOMM_RPN_CMD    0x93
#define BT_RFCOMM_RPN_RSP    0x91
#define BT_RFCOMM_RLS_CMD    0x53
#define BT_RFCOMM_RLS_RSP    0x51
#define BT_RFCOMM_NSC_RSP    0x11

// FCS calc 
#define BT_RFCOMM_CODE_WORD         0xE0 // pol = x8+x2+x1+1
#define BT_RFCOMM_CRC_CHECK_LEN     3
#define BT_RFCOMM_UIHCRC_CHECK_LEN  2
int RFCOMM_CHANNEL_ID = 1;
// used to assemble rfcomm packets
uint8_t rfcomm_out_buffer[1000];
hci_con_handle_t con_handle;

uint16_t source_cid;
                              



//#import "../../RFCOMM/rfcomm.h"

#define INQUIRY_INTERVAL 3

static BTstackManager * btstackManager = nil;

@interface BTstackManager (privat)
- (void)handlePacketWithType:(uint8_t) packet_type forChannel:(uint16_t) channel andData:(uint8_t *)packet withLen:(uint16_t) size;
-(void) ggmHandler:(uint8_t) packet_type forChannel:(uint16_t) channel andData:(uint8_t *) packet 
					withLen:(uint16_t) size ;
- (void)readDeviceInfo;
- (void)storeDeviceInfo;
/* ggm : backward */
-(void) rfcomm_send_packet:(uint16_t) source_cid address:(uint8_t) address 
					control:(uint8_t) control credits:(uint8_t) credits 
					data:(uint8_t*) data len:(uint16_t) len;
-(void) bt_rfcomm_send_sabm:(uint16_t) source_cid initiator:(uint8_t) initiator 
					channel:(uint8_t) channel;
-(void) bt_rfcomm_send_uih_data:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel data:(uint8_t*) data len:(uint16_t) len;
-(void) bt_rfcomm_send_uih_msc_cmd:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel signals:(uint8_t) signals;
-(void) bt_rfcomm_send_uih_pn_command:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel max_frame_size:(uint16_t) max_frame_size;

/* end */
@end

// needed for libBTstack
static void packet_handler(uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size){
	[btstackManager handlePacketWithType:packet_type forChannel:channel andData:packet withLen:size];
}

@implementation BTstackManager

@synthesize delegate = _delegate;
@synthesize deviceInfo;
@synthesize listeners;
@synthesize discoveredDevices;

-(BTstackManager *) init {
	self = [super init];
	if (!self) return self;
	
	state = kDeactivated;
	discoveryState = kInactive;
	connectedToDaemon = NO;
	
	// device discovery
	[self setDiscoveredDevices: [[NSMutableArray alloc] init]];
	
	// delegate and listener
	_delegate = nil;
	[self setListeners:[[NSMutableArray alloc] init]];
	
	// read device database
	[self readDeviceInfo];
	
	// Use Cocoa run loop
	run_loop_init(RUN_LOOP_COCOA);
	
	// our packet handler
	bt_register_packet_handler(packet_handler);
	
	return self;
}

+(BTstackManager *) sharedInstance {
	if (!btstackManager) {
		btstackManager = [[BTstackManager alloc] init];
	}
	return btstackManager;
}

// listeners
-(void) addListener:(id<BTstackManagerListener>)listener{
	[listeners addObject:listener];
}

-(void) removeListener:(id<BTstackManagerListener>)listener{
	[listeners removeObject:listener];
}

// Device info
-(void)readDeviceInfo {
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	NSDictionary * dict = [defaults persistentDomainForName:BTstackManagerID];
	[self setDeviceInfo:[NSMutableDictionary dictionaryWithCapacity:([dict count]+5)]];
	
	// copy entries
	for (id key in dict) {
		NSDictionary *value = [dict objectForKey:key];
		NSMutableDictionary *deviceEntry = [NSMutableDictionary dictionaryWithCapacity:[value count]];
		[deviceEntry addEntriesFromDictionary:value];
		[deviceInfo setObject:deviceEntry forKey:key];
	}
	// NSLog(@"read prefs %@", deviceInfo );
}

-(void)storeDeviceInfo{
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    [defaults setPersistentDomain:deviceInfo forName:BTstackManagerID];
    [defaults synchronize];
	// NSLog(@"store prefs %@", deviceInfo);
	// NSLog(@"Persistence Domain names %@", [defaults persistentDomainNames]);
}

// send events
-(void) sendActivated {
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(activatedBTstackManager:)]){
			[listener activatedBTstackManager:self];
		}
	}
}
-(void) sendActivationFailed:(BTstackError)error {
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(btstackManager:activationFailed:)]){
			[listener btstackManager:self activationFailed:error];
		}
	}
}
-(void) sendDeactivated {
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(deactivatedBTstackManager:)]){
			[listener deactivatedBTstackManager:self];
		}
	}
}
-(void) sendDiscoveryStoppedEvent {
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(discoveryStoppedBTstackManager:)]){
			[listener discoveryStoppedBTstackManager:self];
		}
	}
}
-(void) sendDiscoveryInquiry{
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(discoveryInquiryBTstackManager:)]){
			[listener discoveryInquiryBTstackManager:self];
		}
	}
}
-(void) sendDiscoveryQueryRemoteName:(int)index {
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(btstackManager:discoveryQueryRemoteName:)]){
			[listener btstackManager:self discoveryQueryRemoteName:index];
		}
	}
}
-(void) sendDeviceInfo:(BTDevice*) device{
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:@selector(btstackManager:deviceInfo:)]){
			[listener btstackManager:self deviceInfo:device];
		}
	}
}


// Activation
-(BTstackError) activate {
	
	BTstackError err = 0;
	if (!connectedToDaemon) {
		err = bt_open();
		if (err) return BTSTACK_CONNECTION_TO_BTDAEMON_FAILED;
	}
	connectedToDaemon = YES;
	
	// check system BT
	state = kW4SysBTState;
	bt_send_cmd(&btstack_get_system_bluetooth_enabled);
	
	return err;
}

-(BTstackError) deactivate {
	if (!connectedToDaemon) return BTSTACK_CONNECTION_TO_BTDAEMON_FAILED;
	state = kW4Deactivated;
	bt_send_cmd(&btstack_set_power_mode, HCI_POWER_OFF);
	return 0;
}

-(BOOL) isActive {
	return state == kActivated;
}

-(BOOL) isActivating {
	switch (state){
		case kW4SysBTState:
		case kW4SysBTDisabled:
		case kW4Activated:
			return YES;
		default:
			return NO;
	}
}
-(BOOL) isDiscoveryActive {
	return state == kActivated && (discoveryState != kInactive);
}

// Discovery
-(BTstackError) startDiscovery {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	discoveryState = kW4InquiryMode;
	bt_send_cmd(&hci_write_inquiry_mode, 0x01); // with RSSI
	return 0;
};

-(BTstackError) stopDiscovery{
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	switch (discoveryState){
		case kInactive:
			[self sendDiscoveryStoppedEvent];
			break;
		case kW4InquiryMode:
			discoveryState = kW4InquiryModeBeforeStop;
			break;
		case kInquiry:
			discoveryState = kW4InquiryStop;
			bt_send_cmd(&hci_inquiry_cancel);
			break;
		case kRemoteName: {
			discoveryState = kW4RemoteNameBeforeStop;
			BTDevice *device = [discoveredDevices objectAtIndex:discoveryDeviceIndex];
			bt_send_cmd(&hci_remote_name_request_cancel, [device address]);
			break;
		}
		default:
			NSLog(@"[BTstackManager stopDiscovery] invalid discoveryState %u", discoveryState);
			[self sendDiscoveryStoppedEvent];
			break;
	}
	return 0;
};

-(int) numberOfDevicesFound{
	return [discoveredDevices count];
};

-(BTDevice*) deviceAtIndex:(int)index{
	return (BTDevice*) [discoveredDevices objectAtIndex:index];
};

-(BTDevice*) deviceForAddress:(bd_addr_t*) address{
	for (BTDevice *device in discoveredDevices){
		// NSLog(@"compare %@ to %@", [BTDevice stringForAddress:address], [device addressString]); 
		if ( BD_ADDR_CMP(address, [device address]) == 0){
			return device;
		}
	}
	return nil;
}

- (void) activationHandleEvent:(uint8_t *)packet withLen:(uint16_t) size {
	switch (state) {
						
		case kW4SysBTState:
		case kW4SysBTDisabled:
			
			// BTSTACK_EVENT_SYSTEM_BLUETOOTH_ENABLED
			if ( packet[0] == BTSTACK_EVENT_SYSTEM_BLUETOOTH_ENABLED){
				if (packet[2]){
					// system bt on - first time try to disable it
					if ( state == kW4SysBTState) {
						if (_delegate == nil
							|| ![_delegate respondsToSelector:@selector(disableSystemBluetoothBTstackManager:)]
							|| [_delegate disableSystemBluetoothBTstackManager:self]){
							state = kW4SysBTDisabled;
							bt_send_cmd(&btstack_set_system_bluetooth_enabled, 0);
						} else {
							state = kDeactivated;
							[self sendActivationFailed:BTSTACK_ACTIVATION_FAILED_SYSTEM_BLUETOOTH];
						}
					} else {
						state = kDeactivated;
						[self sendActivationFailed:BTSTACK_ACTIVATION_FAILED_UNKNOWN];
					}
				} else {
					state = kW4Activated;
					bt_send_cmd(&btstack_set_power_mode, HCI_POWER_ON);
				}
			}
			break;
			
		case kW4Activated:
			switch (packet[0]){
				case BTSTACK_EVENT_STATE:
					if (packet[2] == HCI_STATE_WORKING) {
						state = kActivated;
						[self sendActivated];
					}
					break;
				case BTSTACK_EVENT_POWERON_FAILED:
					state = kDeactivated;
					[self sendActivationFailed:BTSTACK_ACTIVATION_POWERON_FAILED];
					break;
					
				default:
					break;
			}
			break;
			
		case kW4Deactivated:
			if (packet[0] == BTSTACK_EVENT_STATE){
				if (packet[2] == HCI_STATE_OFF){
					state = kDeactivated;
					[self sendDeactivated];
				}
			}
			break;
			
		default:
			break;
	}
}

-(void) discoveryRemoteName{
	BOOL found = NO;
	while ( discoveryDeviceIndex < [discoveredDevices count]){
		BTDevice *device = [discoveredDevices objectAtIndex:discoveryDeviceIndex];
		if (device.name) {
			discoveryDeviceIndex ++;
			continue;
		}
		bt_send_cmd(&hci_remote_name_request, [device address], device.pageScanRepetitionMode,
					0, device.clockOffset | 0x8000);
		[self sendDiscoveryQueryRemoteName:discoveryDeviceIndex];
		found = YES;
		break;
	}
	if (!found) {
		// printf("Queried all devices, restart.\n");
		discoveryState = kInquiry;
		bt_send_cmd(&hci_inquiry, HCI_INQUIRY_LAP, INQUIRY_INTERVAL, 0);
		[self sendDiscoveryInquiry];
	}
}

// C'est activated et handler_packet suivant
-(void) discoveryHandleEvent:(uint8_t *)packet withLen:(uint16_t) size {
	bd_addr_t addr;
	int i;
	int numResponses;
	
	switch (discoveryState) {
		case kInactive:
			break;
			
		case kW4InquiryMode:
			if (packet[0] == HCI_EVENT_COMMAND_COMPLETE){
				if (COMMAND_COMPLETE_EVENT(packet, hci_write_inquiry_mode) ) {
					discoveryState = kInquiry;
					bt_send_cmd(&hci_inquiry, HCI_INQUIRY_LAP, INQUIRY_INTERVAL, 0);
					[self sendDiscoveryInquiry];
				}
			}
			break;
		
		case kInquiry:
			
			switch (packet[0]){
				case HCI_EVENT_INQUIRY_RESULT:
					numResponses = packet[2];
					for (i=0; i<numResponses ; i++){
						bt_flip_addr(addr, &packet[3+i*6]);
						// NSLog(@"found %@", [BTDevice stringForAddress:&addr]);
						BTDevice* device = [self deviceForAddress:&addr];
						if (!device) {
							device = [[BTDevice alloc] init];
							[discoveredDevices addObject:device];
							[device setAddress:&addr];
							// get name from deviceInfo
							NSString *addrString = [[device addressString] retain];
							NSMutableDictionary * deviceDict = [deviceInfo objectForKey:addrString];
							[addrString release];
							if (deviceDict){
								device.name = [deviceDict objectForKey:PREFS_REMOTE_NAME];
							}
						}
						// update
						device.pageScanRepetitionMode =   packet [3 + numResponses*(6)         + i*1];
						device.classOfDevice = READ_BT_24(packet, 3 + numResponses*(6+1+1+1)   + i*3);
						device.clockOffset =   READ_BT_16(packet, 3 + numResponses*(6+1+1+1+3) + i*2) & 0x7fff;
						device.rssi  = 0;
						[self sendDeviceInfo:device];
					}
					break;
					
				case HCI_EVENT_INQUIRY_RESULT_WITH_RSSI:
					numResponses = packet[2];
					for (i=0; i<numResponses ;i++){
						bt_flip_addr(addr, &packet[3+i*6]);
						// NSLog(@"found %@", [BTDevice stringForAddress:&addr]);
						BTDevice* device = [self deviceForAddress:&addr];
						if (!device) {
							device = [[BTDevice alloc] init];
							[discoveredDevices addObject:device];
							[device setAddress:&addr];
							// get name from deviceInfo
							NSString *addrString = [[device addressString] retain];
							NSMutableDictionary * deviceDict = [deviceInfo objectForKey:addrString];
							[addrString release];
							if (deviceDict){
								device.name = [deviceDict objectForKey:PREFS_REMOTE_NAME];
							}
						}
						device.pageScanRepetitionMode =   packet [3 + numResponses*(6)         + i*1];
						device.classOfDevice = READ_BT_24(packet, 3 + numResponses*(6+1+1)     + i*3);
						device.clockOffset =   READ_BT_16(packet, 3 + numResponses*(6+1+1+3)   + i*2) & 0x7fff;
						device.rssi  =                    packet [3 + numResponses*(6+1+1+3+2) + i*1];
						[self sendDeviceInfo:device];
					}
					break;

				case HCI_EVENT_INQUIRY_COMPLETE:
					// printf("Inquiry scan done.\n");
					discoveryState = kRemoteName;
					discoveryDeviceIndex = 0;
					[self discoveryRemoteName];
					break;
			}
			break;
			
		case kRemoteName:
			if (packet[0] == HCI_EVENT_REMOTE_NAME_REQUEST_COMPLETE){
				bt_flip_addr(addr, &packet[3]);
				// NSLog(@"Get remote name done for %@", [BTDevice stringForAddress:&addr]);
				BTDevice* device = [self deviceForAddress:&addr];
				if (device) {
					if (packet[2] == 0) {
						// get lenght: first null byte or max 248 chars
						int nameLen = 0;
						while (nameLen < 248 && packet[9+nameLen]) nameLen++;
						NSString *name = [[NSString alloc] initWithBytes:&packet[9] length:nameLen encoding:NSUTF8StringEncoding];
 						device.name = name;
						// set in device info
						NSString *addrString = [[device addressString] retain];
						NSMutableDictionary * deviceDict = [deviceInfo objectForKey:addrString];
						if (!deviceDict){
							deviceDict = [NSMutableDictionary dictionaryWithCapacity:3];
							[deviceInfo setObject:deviceDict forKey:addrString];
						}
						[deviceDict setObject:name forKey:PREFS_REMOTE_NAME];						
						[addrString release];
						[self sendDeviceInfo:device];
					}
					discoveryDeviceIndex++;
					[self discoveryRemoteName];
				}
			}
			break;

		case kW4InquiryModeBeforeStop:
			if (packet[0] == HCI_EVENT_COMMAND_COMPLETE && COMMAND_COMPLETE_EVENT(packet, hci_write_inquiry_mode) ) {
				discoveryState = kInactive;
				[self sendDiscoveryStoppedEvent];
			}
			break;
			
		case kW4InquiryStop:
			if ((packet[0] == HCI_EVENT_INQUIRY_COMPLETE)
			||	packet[0] == HCI_EVENT_COMMAND_COMPLETE && COMMAND_COMPLETE_EVENT(packet, hci_inquiry_cancel)) {
				discoveryState = kInactive;
				[self sendDiscoveryStoppedEvent];
			}
			break;
			
		case kW4RemoteNameBeforeStop:
			if (packet[0] == HCI_EVENT_REMOTE_NAME_REQUEST_COMPLETE
			||  packet[0] == HCI_EVENT_COMMAND_COMPLETE && COMMAND_COMPLETE_EVENT(packet, hci_remote_name_request_cancel)){
				discoveryState = kInactive;
				[self sendDiscoveryStoppedEvent];
			}
			break;
			
		default:
			break;
	}
}

-(void) handleLinkKeyRequestEvent:(uint8_t *)packet withLen:(uint16_t) len {
	bd_addr_t event_addr;
	bt_flip_addr(event_addr, &packet[2]);
	// get link key from deviceInfo
	NSString *devAddress = devAddress = [BTDevice stringForAddress:&event_addr];
	NSMutableDictionary * deviceDict = [deviceInfo objectForKey:devAddress];
	NSData *linkKey = nil;
	if (deviceDict){
		linkKey = [deviceDict objectForKey:PREFS_LINK_KEY];
	}
	if (linkKey) {
		// NSLog(@"Sending link key for %@, value %@", devAddress, linkKey);
		bt_send_cmd(&hci_link_key_request_reply, &event_addr, [linkKey bytes]);
	} else {
		bt_send_cmd(&hci_link_key_request_negative_reply, &event_addr);
	}
}

-(void) handleLinkKeyNotificationEvent:(uint8_t *)packet withLen:(uint16_t) len {
	bd_addr_t event_addr;
	bt_flip_addr(event_addr, &packet[2]); 
	NSString *devAddress = [BTDevice stringForAddress:&event_addr];
	NSData *linkKey = [NSData dataWithBytes:&packet[8] length:16];
	NSMutableDictionary * deviceDict = [deviceInfo objectForKey:devAddress];
	if (!deviceDict){
		deviceDict = [NSMutableDictionary dictionaryWithCapacity:3];
		[deviceInfo setObject:deviceDict forKey:devAddress];
	}
	[deviceDict setObject:linkKey forKey:PREFS_LINK_KEY];
	// NSLog(@"Adding link key for %@, value %@", devAddress, linkKey);
}

-(void) dropLinkKeyForAddress:(bd_addr_t*) address {
	NSString *devAddress = [BTDevice stringForAddress:address];
	NSMutableDictionary * deviceDict = [deviceInfo objectForKey:devAddress];
	[deviceDict removeObjectForKey:PREFS_LINK_KEY];
	// NSLog(@"Removing link key for %@", devAddress);
}

/* Remarque :
 * Se mefier de certains comportements non pr\'evus genre pendant
 * le discovery
 */
-(void) ggmHandler:(uint8_t) packet_type forChannel:(uint16_t) channel andData:(uint8_t *) packet 
					withLen:(uint16_t) size {
    static uint8_t msc_resp_send = 0;
    static uint8_t msc_resp_received = 0;
    static uint8_t credits_used = 0;
    static uint8_t credits_free = 0;
    uint8_t packet_processed = 0;
    bd_addr_t event_addr;
    //if (ready == 0) return;
    switch (packet_type) {
        case L2CAP_DATA_PACKET:
            // rfcomm: data[8] = addr
            // rfcomm: data[9] = command

            //  received 1. message BT_RF_COMM_UA
            if (size == 4 && packet[1] == BT_RFCOMM_UA && packet[0] == 0x03){
                packet_processed++;
                [self bt_rfcomm_send_uih_pn_command:source_cid initiator:1 
													channel:RFCOMM_CHANNEL_ID max_frame_size:100];
            }

            //  received UIH Parameter Negotiation Response
            if (size == 14 && packet[1] == BT_RFCOMM_UIH && packet[3] == BT_RFCOMM_PN_RSP){
                packet_processed++;
                [self bt_rfcomm_send_sabm:source_cid initiator:1 
													channel:RFCOMM_CHANNEL_ID];
            }

            //  received 2. message BT_RF_COMM_UA
            if (size == 4 && packet[1] == BT_RFCOMM_UA && packet[0] == ((RFCOMM_CHANNEL_ID << 3) | 3) ){
                packet_processed++;
                [self bt_rfcomm_send_uih_msc_cmd:source_cid initiator:1 
								channel:RFCOMM_CHANNEL_ID signals:0x8d];// ea=1,fc=0,rtc=1,rtr=1,ic=0,dv=1
            }

            // received BT_RFCOMM_MSC_CMD
            if (size == 8 && packet[1] == BT_RFCOMM_UIH && packet[3] == BT_RFCOMM_MSC_CMD){
                packet_processed++;
                // fine with this
                uint8_t address = packet[0] | 2; // set response 
                packet[3]  = BT_RFCOMM_MSC_RSP;  //  "      "
                [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_UIH
										credits:0x30 data:(uint8_t*)&packet[3] len:4];
                msc_resp_send = 1;
            }

            // received BT_RFCOMM_MSC_RSP
            if (size == 8 && packet[1] == BT_RFCOMM_UIH && packet[3] == BT_RFCOMM_MSC_RSP){
                packet_processed++;
                msc_resp_received = 1;
				for (NSObject<BTstackManagerListener>* listener in listeners) {
					if ([listener respondsToSelector:
							@selector(rfcommConnectionCreatedAtAddress:forChannel:asID:)]){
						[listener rfcommConnectionCreatedAtAddress:connAddr 
							forChannel:connChan asID:0];
					}
				}
            }


            if (packet[1] == BT_RFCOMM_UIH && packet[0] == ((RFCOMM_CHANNEL_ID<<3)|1)){
                packet_processed++;
                credits_used++;
                int length = size-4;
                uint8_t tmp[256];
                memcpy(tmp, packet+3,length);
                tmp[length]='\0';
				for (NSObject<BTstackManagerListener>* listener in listeners) {
					if ([listener respondsToSelector:
							@selector(rfcommDataReceivedForConnectionID: withData: ofLen:)]){
						[listener rfcommDataReceivedForConnectionID:0 withData:tmp	ofLen:length];
					}
				}
            }

            if (packet[1] == BT_RFCOMM_UIH_PF && packet[0] == ((RFCOMM_CHANNEL_ID<<3)|1)){
                packet_processed++;
                credits_used++;
                if (!credits_free) {
                }
                credits_free = packet[2];
            }

            uint8_t send_credits_packet = 0;


            if (credits_used > 40 ) {
                send_credits_packet = 1;
                credits_used = 0;
            }

            if (msc_resp_send && msc_resp_received) {
                send_credits_packet = 1;
                msc_resp_send = msc_resp_received = 0;
            }

            if (send_credits_packet) {
                // send 0x30 credits
                uint8_t initiator = 1;
                uint8_t address = (1 << 0) | (initiator << 1) |  (initiator << 1) | (RFCOMM_CHANNEL_ID << 3);
                [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_UIH_PF
										credits:0x30 data:NULL len:0];
            }

            if (!packet_processed){
                // just dump data for now
                hexdump( packet, size );
            }

            break;
        case HCI_EVENT_PACKET:
            switch (packet[0]) {
                case HCI_EVENT_PIN_CODE_REQUEST:
                    // inform about pin code request
                   bt_flip_addr(event_addr, &packet[2]);
                    bt_send_cmd(&hci_pin_code_request_reply, &event_addr, 4, "0000");
                    break;

                case L2CAP_EVENT_CHANNEL_OPENED:
                    // inform about new l2cap connection
                    bt_flip_addr(event_addr, &packet[3]);
                    //uint16_t psm = READ_BT_16(packet, 11);
                    source_cid = READ_BT_16(packet, 13);
                    con_handle = READ_BT_16(packet, 9);
                    if (packet[2] == 0) {
                       print_bd_addr(event_addr);
                		[self bt_rfcomm_send_sabm:source_cid initiator:1 
											channel:0];
                    } else {
                        exit(1);
                    }
                    break;
            case HCI_EVENT_COMMAND_COMPLETE:
                if (COMMAND_COMPLETE_EVENT(packet, l2cap_create_channel)){
                }
                break;

                default:
                    break;
            }
            break;
        default:
            break;
}

}
// appelle depuis packet_handler
-(void) handlePacketWithType:(uint8_t)packet_type forChannel:(uint16_t)channel andData:(uint8_t *)packet withLen:(uint16_t) size {
	switch (state) {
			
		case kDeactivated:
			break;
		
		// Activation
		case kW4SysBTState:
		case kW4SysBTDisabled:
		case kW4Activated:
		case kW4Deactivated:
			if (packet_type != HCI_EVENT_PACKET) break;
			[self activationHandleEvent:packet withLen:size];
			break;
		
		// Pairing + Discovery
		case kActivated:
			if (packet_type == HCI_EVENT_PACKET) {
			switch (packet[0]){
					break;
				case HCI_EVENT_LINK_KEY_REQUEST:
					[self handleLinkKeyRequestEvent:packet withLen:size];
					break;
				case HCI_EVENT_LINK_KEY_NOTIFICATION:
					[self handleLinkKeyNotificationEvent:packet withLen:size];
					break;
				default:
					break;
			}
			}
			[self ggmHandler:packet_type forChannel:channel andData:packet withLen:size];
			[self discoveryHandleEvent:packet withLen:size];
			break;
		
		default:
			break;
	}
	if ([_delegate respondsToSelector:@selector(btstackManager:handlePacketWithType:forChannel:andData:withLen:)]){
		[_delegate btstackManager:self handlePacketWithType:packet_type forChannel:channel andData:packet withLen:size];
	}
}

// Connections
-(BTstackError) createL2CAPChannelAtAddress:(bd_addr_t*) address withPSM:(uint16_t)psm authenticated:(BOOL)authentication {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	if (state != kActivated) return BTSTACK_BUSY;
#if 0
	// ...f (state 
	// store params
	connType = 0;
	BD_ADDR_COPY(&connAddr, address);
	connPSM = psm;
	connAuth = authentication;
	
	// send write authentication enabled
	bt_send_cmd(&hci_write_authentication_enable, 0);	
	state = kW4AuthenticationEnableCommand;
#endif
	return 0;
};
-(BTstackError) sendL2CAPPacketForChannelID:(uint16_t)channelID {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	return 0;
};
-(BTstackError) closeL2CAPChannelWithID:(uint16_t) channelID {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	return 0;
};

-(BTstackError) createRFCOMMConnectionAtAddress:(bd_addr_t*) address withChannel:(uint16_t)channel authenticated:(BOOL)authentication {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	if (state != kActivated) return BTSTACK_BUSY;
	for (NSObject<BTstackManagerListener>* listener in listeners) {
		if ([listener respondsToSelector:
			@selector(rfcommConnectionCreatedAtAddress:forChannel:asID:)]){
			[listener rfcommConnectionCreatedAtAddress:connAddr 
				forChannel:connChan asID:1];
		}
	}
//#if 0
	// store params
	connType = 1;
	BD_ADDR_COPY(&connAddr, address);
	connChan = channel;
	connAuth = authentication;
	bt_send_cmd(&l2cap_create_channel, address, 0x03);	
	state =kActivated;
//#endif 
	return 0;
};
-(BTstackError) sendRFCOMMPacketForChannelID:(uint16_t)connectionID {
	if (state < kActivated) return BTSTACK_NOT_ACTIVATED;
	return 0;
};
-(BTstackError) closeRFCOMMConnectionWithID:(uint16_t) connectionID {
	if (state <kActivated) return BTSTACK_NOT_ACTIVATED;
	return 0;
};

/* GGM : backward compat */

-(void) rfcomm_send_packet:(uint16_t) source_cid address:(uint8_t) address 
					control:(uint8_t) control credits:(uint8_t) credits 
					data:(uint8_t*) data len:(uint16_t) len
{

    uint16_t pos = 0;
    uint8_t crc_fields = 3;

    rfcomm_out_buffer[pos++] = address;
    rfcomm_out_buffer[pos++] = control;

    // length field can be 1 or 2 octets
    if (len < 128){
        rfcomm_out_buffer[pos++] = (len << 1)| 1;     // bits 0-6
    } else {
        rfcomm_out_buffer[pos++] = (len & 0x7f) << 1; // bits 0-6
        rfcomm_out_buffer[pos++] = len >> 7;          // bits 7-14
        crc_fields++;
    }

    // add credits for UIH frames when PF bit is set
    if (control == BT_RFCOMM_UIH_PF){
        rfcomm_out_buffer[pos++] = credits;
    }

    // copy actual data
    memcpy(&rfcomm_out_buffer[pos], data, len);
    pos += len;

    // UIH frames only calc FCS over address + control (5.1.1)
    if ((control & 0xef) == BT_RFCOMM_UIH){
        crc_fields = 2;
    }
    rfcomm_out_buffer[pos++] =  crc8_calc(rfcomm_out_buffer, crc_fields); // calc fcs
    bt_send_l2cap( source_cid, rfcomm_out_buffer, pos);
}

-(void) bt_rfcomm_send_sabm:(uint16_t) source_cid initiator:(uint8_t) initiator 
					channel:(uint8_t) channel
{
    uint8_t address = (1 << 0) | (initiator << 1) |  (initiator << 1) | (channel << 3);
    [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_SABM
							credits:0x0 data:NULL len:0];
}

-(void) bt_rfcomm_send_uih_data:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel data:(uint8_t*) data len:(uint16_t) len
{
    uint8_t address = (1 << 0) | (initiator << 1) |  (initiator << 1) | (channel << 3);
    [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_UIH
							credits:0x0 data:data len:len];
}

-(void) bt_rfcomm_send_uih_msc_cmd:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel signals:(uint8_t) signals
{
    uint8_t address = (1 << 0) | (initiator << 1); // EA and C/R bit set - always server channel 0
    uint8_t payload[4];
    uint8_t pos = 0;
    payload[pos++] = BT_RFCOMM_MSC_CMD;
    payload[pos++] = 2 << 1 | 1;  // len
    payload[pos++] = (1 << 0) | (1 << 1) | (0 << 2) | (channel << 3); // shouldn't D = initiator = 1 ?
    payload[pos++] = signals;
    [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_UIH
							credits:0x0 data:(uint8_t *)payload len:pos];
}
-(void) bt_rfcomm_send_uih_pn_command:(uint16_t) source_cid initiator:(uint8_t) initiator
					channel:(uint8_t) channel max_frame_size:(uint16_t) max_frame_size
{
    uint8_t payload[10];
    uint8_t address = (1 << 0) | (initiator << 1); // EA and C/R bit set - always server channel 0
    uint8_t pos = 0;
    payload[pos++] = BT_RFCOMM_PN_CMD;
    payload[pos++] = 8 << 1 | 1;  // len
    payload[pos++] = channel << 1;
    payload[pos++] = 0xf0; // pre defined for Bluetooth, see 5.5.3 of TS 07.10 Adaption for RFCOMM
    payload[pos++] = 0; // priority
    payload[pos++] = 0; // max 60 seconds ack
    payload[pos++] = max_frame_size & 0xff; // max framesize low
    payload[pos++] = max_frame_size >> 8;   // max framesize high
    payload[pos++] = 0x00; // number of retransmissions
    payload[pos++] = 0x00; // unused error recovery window
    [self rfcomm_send_packet:source_cid address:address control:BT_RFCOMM_UIH
							credits:0x0 data:(uint8_t *)payload len:pos];
}


/* GGM : end of backward compat */




@end
